/* Othello.java */

import java.applet.*;
import java.awt.*;
import java.awt.event.*;

/** Othello. */
public class Othello
	extends Applet
	implements MouseListener, Runnable
{
	/** Version string. */
	private static final String version="0.9 beta";
	
	/** Applet mode. */
	private int mode;

	/** Mode "Nothing". */
	private final int NOTHING  = 0;
	/** Mode "Menu". */
	private final int MENU     = 1;
	/** Mode "Playing". */
	private final int PLAYING  = 2;
	/** Mode "Game Over". */
	private final int GAMEOVER = 3;
	
	/** Background image. */
	private Image background;
	/** White chip image. */
	private Image white;
	/** Black chip image. */
	private Image black;
	/** Diode images (on/off). */
	private Image [] light=new Image[2];
	/** Button images (on/off). */
	private Image [] button=new Image[2];
	/** LCD-style image handler. */
	private OthelloDigits digi;
	
	/** Gameboard. */
	private OthelloBoard board;
	/** Players. */
	private OthelloPlayer player[]=new OthelloPlayer[2];
	/** Current player. */
	private int currentPlayer;
	
	/** Blinkes diodes. */
	private Thread blinker;
	/** Diode state. */
	private int blink=0;
	/** Game completion state.
	    A value of 1 means that Player 1 (black) lost, 2 that Player 2 /(white) lost, and 3 implies a draw. */
	private int loser=3;
	
	/** Button positions. */
	private Point [] btnPos={new Point(430, 170), new Point(430, 200)};
	/** Button width. */
	private int btnWidth=18;
	/** Button height. */
	private int btnHeight=14;
	/** Button states. */
	private boolean [] btnDown={false, false};
	
	/** States which of the players are controlled by the computer. */
	private boolean [] ai={false, true};
	
	/** Updates screen.
	    This method is believed to cause smoother graphics. */
	public void update(Graphics g)
	{
		paint(g);
	}
	
	/** Runs thread. */
	public void run()
	{
		Thread thisThread=Thread.currentThread();
		if(thisThread==blinker)
		{
			while(blinker!=null)
			{
				blink=1-blink;
				repaint(352, 247, 10, 58);
				//repaint();
				try{Thread.sleep(100);}
				catch(InterruptedException e) {}
			}
			return;
		}
	}
	
	/** Updates screen. */
	public void paint(Graphics graphics)
	{
		// Set up off-screen Graphics object.
		Image virtualScreen=createImage(480, 352);
		Graphics g=virtualScreen.getGraphics();
		
		// Draw off-screen
		switch(mode)
		{
			case NOTHING:
				break;
			
			case MENU:
				// Background
				g.drawImage(background, 0, 0, this);
				
				// Numbers
				digi.drawDigit(ai[0]?'A':'H', g, 424, 240);
				digi.drawDigit(ai[0]?'I':'U', g, 440, 240);
				digi.drawDigit(ai[1]?'A':'H', g, 424, 288);
				digi.drawDigit(ai[1]?'I':'U', g, 440, 288);
				
				// Lights
				g.drawImage(light[1], 352, 247, this);
				g.drawImage(light[1], 352, 295, this);
				
				// Buttons
				for(int n=0;n<btnDown.length;n++)
					g.drawImage(button[btnDown[n]?1:0], btnPos[n].x, btnPos[n].y, this);
				
				break;
			
			case PLAYING:
				// Background
				g.drawImage(background, 0, 0, this);
				
				// Chips
				for(int i=0;i<board.getWidth();i++)
					for(int j=0;j<board.getHeight();j++)
						switch(board.get(i, j))
						{
							case board.BLACK:
								g.drawImage(black, 16+i*40, 16+j*40, this);
								break;
							case board.WHITE:
								g.drawImage(white, 16+i*40, 16+j*40, this);
								break;
						}
				
				// Numbers
				digi.drawNumber(board.getNumber(board.BLACK), g, 424, 240);
				digi.drawNumber(board.getNumber(board.WHITE), g, 424, 288);
				
				// Lights
				g.drawImage(light[currentPlayer], 352, 247, this);
				g.drawImage(light[1-currentPlayer], 352, 295, this);
				
				// Buttons
				for(int n=0;n<btnDown.length;n++)
					g.drawImage(button[btnDown[n]?1:0], btnPos[n].x, btnPos[n].y, this);
				
				break;
			
			case GAMEOVER:
				// Background
				g.drawImage(background, 0, 0, this);
				
				// Chips
				for(int i=0;i<board.getWidth();i++)
					for(int j=0;j<board.getHeight();j++)
						switch(board.get(i, j))
						{
							case board.BLACK:
								g.drawImage(black, 16+i*40, 16+j*40, this);
								break;
							case board.WHITE:
								g.drawImage(white, 16+i*40, 16+j*40, this);
								break;
						}
				
				// Numbers
				digi.drawNumber(board.getNumber(board.BLACK), g, 424, 240);
				digi.drawNumber(board.getNumber(board.WHITE), g, 424, 288);
				
				// Lights
				g.drawImage(light[(loser&1)!=0?1:blink], 352, 247, this);
				g.drawImage(light[(loser&2)!=0?1:1-blink], 352, 295, this);
				
				// Buttons
				for(int n=0;n<btnDown.length;n++)
					g.drawImage(button[btnDown[n]?1:0], btnPos[n].x, btnPos[n].y, this);
				
				break;
		}
		// Update screen
		graphics.drawImage(virtualScreen, 0, 0, this);
		virtualScreen.flush();
		g.dispose();
	}
	
	/** Startup. */
	public void init()
	{
		mode=NOTHING;
		
		// Load images
		background= getImage(getCodeBase(), "pix/base.gif");
		black=      getImage(getCodeBase(), "pix/schwartz.gif");
		white=      getImage(getCodeBase(), "pix/weiss.gif");
		light[0]=   getImage(getCodeBase(), "pix/lighton.gif");
		light[1]=   getImage(getCodeBase(), "pix/lightoff.gif");
		button[0]=  getImage(getCodeBase(), "pix/buttonup.gif");
		button[1]=  getImage(getCodeBase(), "pix/buttondown.gif");
		
		MediaTracker tracker=new MediaTracker(this);
        tracker.addImage(background, 0);
        tracker.addImage(black,      1);
        tracker.addImage(white,      2);
        tracker.addImage(light[0],   3);
        tracker.addImage(light[1],   4);
        tracker.addImage(button[0],  5);
        tracker.addImage(button[1],  6);
        try{tracker.waitForAll();}
        catch(InterruptedException e){};
        
        // LCD handler
		digi=new OthelloDigits(this);
		
		addMouseListener(this);
		mode=MENU;
	}
	
	/** Resets game state.
	    The applet will return to "Menu" mode. */
	private void reset()
	{
		mode=NOTHING;
		player[0].stopMove();
		player[1].stopMove();
		mode=MENU;
		repaint();
	}
	
	/** Start a new game of Othello. */
	private void startNewGame()
	{
		// Turn off threads
		blinker=null;
		//animator=null;
		
		board=new OthelloBoard();
		player[0]=ai[0]?(OthelloPlayer)new OthelloAI(board.BLACK, 4, this):(OthelloPlayer)new OthelloHuman(board.BLACK, this);
		player[1]=ai[1]?(OthelloPlayer)new OthelloAI(board.WHITE, 4, this):(OthelloPlayer)new OthelloHuman(board.WHITE, this);
		currentPlayer=0;
		player[currentPlayer].startMove(board);
		
		mode=PLAYING;
	}
	
	/** Sets next player as current player. */
	private void setNextPlayer()
	{
		currentPlayer=1-currentPlayer;
	}
	
	/** Processes clicks on "Mode"-button.
	    This method will change the player composition,
	    such that the sequence (HU/HU, HU/AI, AI/HU, AI/AI) is acheived. */
	private void setNextPlayMode()
	{
		ai[0]=ai[0]^ai[1];
		ai[1]=!ai[1];
	}
	
	/** This method is called by players who want to make their move.
	    It will return false if the move was invalid.
	    If the move was successful, true is returned,
	    and the next player is informed ("started"). */
	public boolean tryToSet(int x, int y)
	{
		if(mode==PLAYING)
		{
			if(board.canPut(x, y, player[currentPlayer].getColor()))
			{
				board.put(x, y, player[currentPlayer].getColor());
				repaint();
				
				setNextPlayer();
				Point [] p=board.canPut(player[currentPlayer].getColor());
				if(p.length<1)
				{
					setNextPlayer();
					p=board.canPut(player[currentPlayer].getColor());
					if(p.length<1)
					{
						setGameOver();
						return true;
					}
				}
				player[currentPlayer].startMove(board);
				return true;
			}
		}
		return false;
	}
	
	/** Sets "Game over" mode.
	    This will start the diode blink-effect. */
	private void setGameOver()
	{
		mode=NOTHING;
		int black=board.getNumber(board.BLACK);
		int white=board.getNumber(board.WHITE);
		loser=3;
		if(black>white)
			loser=2;
		else if(white>black)
			loser=1;
		mode=GAMEOVER;
		blinker=new Thread(this);
		blinker.start();
	}
	
	/** Returns version string. */
	public static String getVersion()
	{
		return version;
	}
	
	/** Derived from MouseListener. */
	public void mouseClicked(MouseEvent e) {}
	
	/** Derived from MouseListener. */
	public void mouseEntered(MouseEvent e) {}
	
	/** Derived from MouseListener. */
	public void mouseExited(MouseEvent e) {}
	
	/** Derived from MouseListener.
	    Will cause buttons to enter "down" mode. */
	public void mousePressed(MouseEvent e)
	{
		Point p=e.getPoint();
		for(int n=0;n<btnDown.length;n++)
			if((p.x>=btnPos[n].x) && (p.x<btnPos[n].x+btnWidth) && (p.y>=btnPos[n].y) && (p.y<btnPos[n].y+btnHeight))
				btnDown[n]=true;
		repaint();
	}
	
	/** Derived from MouseListener.
	    Will cause buttons to exit "down" mode,
	    and executes issueButtonClick(). */
	public void mouseReleased(MouseEvent e)
	{
		Point p=e.getPoint();
		for(int n=0;n<btnDown.length;n++)
			if((p.x>=btnPos[n].x) && (p.x<btnPos[n].x+btnWidth) && (p.y>=btnPos[n].y) && (p.y<btnPos[n].y+btnHeight))
				if(btnDown[n])
					issueButtonClick(n);
		for(int n=0;n<btnDown.length;n++)
			btnDown[n]=false;
		repaint();
	}
	
	/** This method will respond to a button click. */
	private void issueButtonClick(int btn)
	{
		switch(mode)
		{
			case MENU:
				switch(btn)
				{
					case 0:
						startNewGame();
						break;
					case 1:
						setNextPlayMode();
						break;
				}
				break;
			case PLAYING:
			case GAMEOVER:
				switch(btn)
				{
					case 0:
						reset();
						break;
				}
				break;
		}
		repaint();
	}
}
